@@ -10,6 +10,8 @@
 #include <signal.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <cstdint>
+#include <ctime>
 #include <memory>
 #include <stdexcept>
 #include <unordered_map>
@@ -22,6 +24,7 @@
 #include "fcitx-utils/eventdispatcher.h"
 #include "fcitx-utils/i18n.h"
 #include "fcitx-utils/log.h"
+#include "fcitx-utils/macros.h"
 #include "fcitx-utils/misc.h"
 #include "fcitx-utils/standardpath.h"
 #include "fcitx-utils/stringutils.h"
@@ -974,6 +977,13 @@ Instance::Instance(int argc, char **argv) {
 #ifdef ENABLE_KEYBOARD
             if (keyEvent.forward()) {
                 FCITX_D();
+                // Always let the release key go through, since it shouldn't
+                // produce character. Otherwise it may wrongly trigger wayland
+                // client side repetition.
+                if (keyEvent.isRelease()) {
+                    keyEvent.filter();
+                    return;
+                }
                 auto *inputState = ic->propertyFor(&d->inputStateFactory_);
                 if (auto *xkbState = inputState->customXkbState()) {
                     if (auto utf32 = xkb_state_key_get_utf32(
@@ -989,10 +999,8 @@ Instance::Instance(int argc, char **argv) {
                                 keyEvent.origKey().sym()) {
                             return;
                         }
-                        if (!keyEvent.isRelease()) {
-                            FCITX_KEYTRACE() << "Will commit char: " << utf32;
-                            ic->commitString(utf8::UCS4ToUTF8(utf32));
-                        }
+                        FCITX_KEYTRACE() << "Will commit char: " << utf32;
+                        ic->commitString(utf8::UCS4ToUTF8(utf32));
                         keyEvent.filterAndAccept();
                     } else if (!keyEvent.key().states().test(KeyState::Ctrl) &&
                                keyEvent.rawKey().sym() !=
@@ -1356,6 +1364,9 @@ void Instance::handleSignal() {
             exit();
         } else if (signo == SIGUSR1) {
             reloadConfig();
+        } else if (signo == SIGCHLD) {
+            d->zombieReaper_->setNextInterval(2000000);
+            d->zombieReaper_->setOneShot();
         }
     }
 }
@@ -1406,6 +1417,15 @@ void Instance::initialize() {
             }
             return false;
         });
+    d->zombieReaper_ = d->eventLoop_.addTimeEvent(
+        CLOCK_MONOTONIC, now(CLOCK_MONOTONIC), 0,
+        [](EventSourceTime *, uint64_t) {
+            pid_t res;
+            while ((res = waitpid(-1, nullptr, WNOHANG)) > 0) {
+            }
+            return false;
+        });
+    d->zombieReaper_->setEnabled(false);
 
     d->exitEvent_ = d->eventLoop_.addExitEvent([this](EventSource *) {
         FCITX_DEBUG() << "Running save...";
